"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ReactiveDict = void 0;

var _standaloneTracker = require("standalone-tracker");

var _ejson = _interopRequireDefault(require("ejson"));

var _mongoId = _interopRequireDefault(require("./mongoId"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function _typeof(obj) { "@babel/helpers - typeof"; if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

var hasOwn = Object.prototype.hasOwnProperty; // XXX come up with a serialization method which canonicalizes object key
// order, which would allow us to use objects as values for equals.

function stringify(value) {
  if (value === undefined) {
    return 'undefined';
  }

  return _ejson["default"].stringify(value);
}

function parse(serialized) {
  if (serialized === undefined || serialized === 'undefined') {
    return undefined;
  }

  return _ejson["default"].parse(serialized);
}

function changed(v) {
  return v && v.changed();
} // XXX COMPAT WITH 0.9.1 : accept migrationData instead of dictName

/**
 * @class
 * @instanceName ReactiveDict
 * @summary Constructor for a ReactiveDict,
 *  which represents a reactive dictionary of key/value pairs.
 * @locus Client
 * @param {String} [name] Optional.  When a name is passed,
 *  preserves contents across Hot Code Pushes
 * @param {Object} [initialValue] Optional.  The default values for the dictionary
 */
// eslint-disable-next-line import/prefer-default-export


var ReactiveDict = /*#__PURE__*/function () {
  function ReactiveDict(dictName, dictData) {
    _classCallCheck(this, ReactiveDict);

    // this.keys: key -> value
    this.keys = {};
    this._migratedDictData = {}; // name -> data

    this._dictsToMigrate = {}; // name -> ReactiveDict

    if (dictName) {
      // name given; migration will be performed
      if (typeof dictName === 'string') {
        // the normal case, argument is a string name.
        // Only run migration logic on client, it will cause
        // duplicate name errors on server during reloads.
        // _registerDictForMigrate will throw an error on duplicate name.
        if (typeof window !== 'undefined') {
          this._registerDictForMigrate(dictName, this);
        }

        var migratedData = typeof window !== 'undefined' && this._loadMigratedDict(dictName);

        if (migratedData) {
          // Don't stringify migrated data
          this.keys = migratedData;
        } else {
          // Use _setObject to make sure values are stringified
          this._setObject(dictData || {});
        }

        this.name = dictName;
      } else if (_typeof(dictName) === 'object') {
        // back-compat case: dictName is actually migrationData
        // Use _setObject to make sure values are stringified
        this._setObject(dictName);
      } else {
        throw new Error("Invalid ReactiveDict argument: ".concat(dictName));
      }
    } else if (_typeof(dictData) === 'object') {
      this._setObject(dictData);
    }

    this.allDeps = new _standaloneTracker.Tracker.Dependency();
    this.keyDeps = {}; // key -> Dependency

    this.keyValueDeps = {}; // key -> Dependency
  } // set() began as a key/value method, but we are now overloading it
  // to take an object of key/value pairs, similar to backbone
  // http://backbonejs.org/#Model-set

  /**
   * @summary Set a value for a key in the ReactiveDict. Notify any listeners
   * that the value has changed (eg: redraw templates, and rerun any
   * [`Tracker.autorun`](#tracker_autorun) computations, that called
   * [`ReactiveDict.get`](#ReactiveDict_get) on this `key`.)
   * @locus Client
   * @param {String} key The key to set, eg, `selectedItem`
   * @param {EJSONable | undefined} value The new value for `key`
   */


  _createClass(ReactiveDict, [{
    key: "set",
    value: function set(keyOrObject, value) {
      if (_typeof(keyOrObject) === 'object' && value === undefined) {
        // Called as `dict.set({...})`
        this._setObject(keyOrObject);

        return;
      } // the input isn't an object, so it must be a key
      // and we resume with the rest of the function


      var key = keyOrObject;
      var stringifiedValue = stringify(value);
      var keyExisted = hasOwn.call(this.keys, key);
      var oldSerializedValue = keyExisted ? this.keys[key] : 'undefined';
      var isNewValue = stringifiedValue !== oldSerializedValue;
      this.keys[key] = stringifiedValue;

      if (isNewValue || !keyExisted) {
        // Using the changed utility function here because this.allDeps might not exist yet,
        // when setting initial data from constructor
        changed(this.allDeps);
      } // Don't trigger changes when setting initial data from constructor,
      // this.KeyDeps is undefined in this case


      if (isNewValue && this.keyDeps) {
        changed(this.keyDeps[key]);

        if (this.keyValueDeps[key]) {
          changed(this.keyValueDeps[key][oldSerializedValue]);
          changed(this.keyValueDeps[key][stringifiedValue]);
        }
      }
    }
    /**
     * @summary Set a value for a key if it hasn't been set before.
     * Otherwise works exactly the same as [`ReactiveDict.set`](#ReactiveDict-set).
     * @locus Client
     * @param {String} key The key to set, eg, `selectedItem`
     * @param {EJSONable | undefined} value The new value for `key`
     */

  }, {
    key: "setDefault",
    value: function setDefault(keyOrObject, value) {
      if (_typeof(keyOrObject) === 'object' && value === undefined) {
        // Called as `dict.setDefault({...})`
        this._setDefaultObject(keyOrObject);

        return;
      } // the input isn't an object, so it must be a key
      // and we resume with the rest of the function


      var key = keyOrObject;

      if (!hasOwn.call(this.keys, key)) {
        this.set(key, value);
      }
    }
    /**
     * @summary Get the value assiciated with a key. If inside a [reactive
     * computation](#reactivity), invalidate the computation the next time the
     * value associated with this key is changed by
     * [`ReactiveDict.set`](#ReactiveDict-set).
     * This returns a clone of the value, so if it's an object or an array,
     * mutating the returned value has no effect on the value stored in the
     * ReactiveDict.
     * @locus Client
     * @param {String} key The key of the element to return
     */

  }, {
    key: "get",
    value: function get(key) {
      this._ensureKey(key);

      this.keyDeps[key].depend();
      return parse(this.keys[key]);
    }
    /**
     * @summary Test if the stored entry for a key is equal to a value. If inside a
     * [reactive computation](#reactivity), invalidate the computation the next
     * time the variable changes to or from the value.
     * @locus Client
     * @param {String} key The name of the session variable to test
     * @param {String | Number | Boolean | null | undefined} value The value to
     * test against
     */

  }, {
    key: "equals",
    value: function equals(key, value) {
      var _this = this;

      // We don't allow objects (or arrays that might include objects) for
      // .equals, because JSON.stringify doesn't canonicalize object key
      // order. (We can make equals have the right return value by parsing the
      // current value and using EJSON.equals, but we won't have a canonical
      // element of keyValueDeps[key] to store the dependency.) You can still use
      // "EJSON.equals(reactiveDict.get(key), value)".
      //
      // XXX we could allow arrays as long as we recursively check that there
      // are no objects
      if (typeof value !== 'string' && typeof value !== 'number' && typeof value !== 'boolean' && typeof value !== 'undefined' && !(value instanceof Date) && !(value instanceof _mongoId["default"].ObjectID) && value !== null) {
        throw new Error('ReactiveDict.equals: value must be scalar');
      }

      var serializedValue = stringify(value);

      if (_standaloneTracker.Tracker.active) {
        this._ensureKey(key);

        if (!hasOwn.call(this.keyValueDeps[key], serializedValue)) {
          this.keyValueDeps[key][serializedValue] = new _standaloneTracker.Tracker.Dependency();
        }

        var isNew = this.keyValueDeps[key][serializedValue].depend();

        if (isNew) {
          _standaloneTracker.Tracker.onInvalidate(function () {
            // clean up [key][serializedValue] if it's now empty, so we don't
            // use O(n) memory for n = values seen ever
            if (!_this.keyValueDeps[key][serializedValue].hasDependents()) {
              delete _this.keyValueDeps[key][serializedValue];
            }
          });
        }
      }

      var oldValue;

      if (hasOwn.call(this.keys, key)) {
        oldValue = parse(this.keys[key]);
      }

      return _ejson["default"].equals(oldValue, value);
    }
    /**
     * @summary Get all key-value pairs as a plain object. If inside a [reactive
     * computation](#reactivity), invalidate the computation the next time the
     * value associated with any key is changed by
     * [`ReactiveDict.set`](#ReactiveDict-set).
     * This returns a clone of each value, so if it's an object or an array,
     * mutating the returned value has no effect on the value stored in the
     * ReactiveDict.
     * @locus Client
     */

  }, {
    key: "all",
    value: function all() {
      var _this2 = this;

      this.allDeps.depend();
      var ret = {};
      Object.keys(this.keys).forEach(function (key) {
        ret[key] = parse(_this2.keys[key]);
      });
      return ret;
    }
    /**
     * @summary remove all key-value pairs from the ReactiveDict. Notify any
     * listeners that the value has changed (eg: redraw templates, and rerun any
     * [`Tracker.autorun`](#tracker_autorun) computations, that called
     * [`ReactiveDict.get`](#ReactiveDict_get) on this `key`.)
     * @locus Client
     */

  }, {
    key: "clear",
    value: function clear() {
      var _this3 = this;

      var oldKeys = this.keys;
      this.keys = {};
      this.allDeps.changed();
      Object.keys(oldKeys).forEach(function (key) {
        changed(_this3.keyDeps[key]);

        if (_this3.keyValueDeps[key]) {
          changed(_this3.keyValueDeps[key][oldKeys[key]]);
          changed(_this3.keyValueDeps[key].undefined);
        }
      });
    }
    /**
     * @summary remove a key-value pair from the ReactiveDict. Notify any listeners
     * that the value has changed (eg: redraw templates, and rerun any
     * [`Tracker.autorun`](#tracker_autorun) computations, that called
     * [`ReactiveDict.get`](#ReactiveDict_get) on this `key`.)
     * @locus Client
     * @param {String} key The key to delete, eg, `selectedItem`
     */

  }, {
    key: "delete",
    value: function _delete(key) {
      var didRemove = false;

      if (hasOwn.call(this.keys, key)) {
        var oldValue = this.keys[key];
        delete this.keys[key];
        changed(this.keyDeps[key]);

        if (this.keyValueDeps[key]) {
          changed(this.keyValueDeps[key][oldValue]);
          changed(this.keyValueDeps[key].undefined);
        }

        this.allDeps.changed();
        didRemove = true;
      }

      return didRemove;
    }
    /**
     * @summary Clear all values from the reactiveDict and prevent it from being
     * migrated on a Hot Code Pushes. Notify any listeners
     * that the value has changed (eg: redraw templates, and rerun any
     * [`Tracker.autorun`](#tracker_autorun) computations, that called
     * [`ReactiveDict.get`](#ReactiveDict_get) on this `key`.)
     * @locus Client
     */

  }, {
    key: "destroy",
    value: function destroy() {
      this.clear();

      if (this.name && hasOwn.call(this._dictsToMigrate, this.name)) {
        delete this._dictsToMigrate[this.name];
      }
    }
  }, {
    key: "_setObject",
    value: function _setObject(object) {
      var _this4 = this;

      Object.keys(object).forEach(function (key) {
        _this4.set(key, object[key]);
      });
    }
  }, {
    key: "_setDefaultObject",
    value: function _setDefaultObject(object) {
      var _this5 = this;

      Object.keys(object).forEach(function (key) {
        _this5.setDefault(key, object[key]);
      });
    }
  }, {
    key: "_ensureKey",
    value: function _ensureKey(key) {
      if (!(key in this.keyDeps)) {
        this.keyDeps[key] = new _standaloneTracker.Tracker.Dependency();
        this.keyValueDeps[key] = {};
      }
    } // Get a JSON value that can be passed to the constructor to
    // create a new ReactiveDict with the same contents as this one

  }, {
    key: "_getMigrationData",
    value: function _getMigrationData() {
      // XXX sanitize and make sure it's JSONible?
      return this.keys;
    }
  }, {
    key: "_loadMigratedDict",
    value: function _loadMigratedDict(dictName) {
      if (hasOwn.call(this._migratedDictData, dictName)) {
        var data = this._migratedDictData[dictName];
        delete this._migratedDictData[dictName];
        return data;
      }

      return null;
    }
  }, {
    key: "_registerDictForMigrate",
    value: function _registerDictForMigrate(dictName, dict) {
      if (hasOwn.call(this._dictsToMigrate, dictName)) {
        throw new Error("Duplicate ReactiveDict name: ".concat(dictName));
      }

      this._dictsToMigrate[dictName] = dict;
    }
  }]);

  return ReactiveDict;
}();

exports.ReactiveDict = ReactiveDict;